'use strict';

var _createClass = require('babel-runtime/helpers/create-class')['default'];

var _classCallCheck = require('babel-runtime/helpers/class-call-check')['default'];

var _regeneratorRuntime = require('babel-runtime/regenerator')['default'];

var _getIterator = require('babel-runtime/core-js/get-iterator')['default'];

var _Map = require('babel-runtime/core-js/map')['default'];

var _interopRequireDefault = require('babel-runtime/helpers/interop-require-default')['default'];

var _interopRequireWildcard = require('babel-runtime/helpers/interop-require-wildcard')['default'];

Object.defineProperty(exports, '__esModule', {
  value: true
});

var _levelgraph = require('levelgraph');

var _levelgraph2 = _interopRequireDefault(_levelgraph);

var _levelgraphN3 = require('levelgraph-n3');

var _levelgraphN32 = _interopRequireDefault(_levelgraphN3);

var _level = require('level');

var _level2 = _interopRequireDefault(_level);

var _lodash = require('lodash');

var _lodash2 = _interopRequireDefault(_lodash);

var _bluebird = require('bluebird');

var _bluebird2 = _interopRequireDefault(_bluebird);

var _helperWikipedia = require('./helper/wikipedia');

var wikipedia = _interopRequireWildcard(_helperWikipedia);

var _helperPermutations = require('./helper/permutations');

var _helperPermutations2 = _interopRequireDefault(_helperPermutations);

var MESH = 'http://id.nlm.nih.gov/mesh/';
var MESHV = 'http://id.nlm.nih.gov/mesh/vocab#';
var RDF = 'http://www.w3.org/1999/02/22-rdf-syntax-ns#';
var RDFS = 'http://www.w3.org/2000/01/rdf-schema#';

var MeshTree = (function () {
  function MeshTree(config) {
    _classCallCheck(this, MeshTree);

    config = config || {};

    var DB = undefined;
    if (config.level) {
      DB = config.level;
    } else {
      var dbPath = config.dbPath || process.env['PATH_TO_MESH_DB'] || 'dbtest';
      DB = (0, _level2['default'])(dbPath);
    }

    this.db = (0, _levelgraphN32['default'])((0, _levelgraph2['default'])(DB));
    this.dbSearch = _bluebird2['default'].promisify(this.db.search, { multiArgs: false });
  }

  /*
  * Returns array of all descriptor records
  *
  * Four types of descriptor records exist:
  * - TopicalDescriptor
  * - GeographicalDescriptor
  * - PublicationType
  * - CheckTag
  *
  * By default, only TopicalDescriptor and GeographicalDescriptor are included
  */

  _createClass(MeshTree, [{
    key: 'formatID',
    value: function formatID(id, format) {
      if (format === 'rdf') {
        if (id.startsWith(MESH)) {
          return id;
        } else {
          return MESH + id;
        }
      } else if (format === 'mesh') {
        if (id.startsWith(MESH)) {
          return id.replace(MESH, '');
        } else {
          return id;
        }
      } else {
        return id;
      }
    }
  }]);

  return MeshTree;
})();

MeshTree.prototype.getAllDescUIs = _bluebird2['default'].coroutine(_regeneratorRuntime.mark(function callee$0$0(opts) {
  var format, classes, results, _iteratorNormalCompletion, _didIteratorError, _iteratorError, _iterator, _step, cl, result;

  return _regeneratorRuntime.wrap(function callee$0$0$(context$1$0) {
    var _this = this;

    while (1) switch (context$1$0.prev = context$1$0.next) {
      case 0:
        opts = opts || {};
        format = opts.format || 'rdf';
        classes = opts.classes || ['TopicalDescriptor', 'GeographicalDescriptor'];
        results = [];
        _iteratorNormalCompletion = true;
        _didIteratorError = false;
        _iteratorError = undefined;
        context$1$0.prev = 7;
        _iterator = _getIterator(classes);

      case 9:
        if (_iteratorNormalCompletion = (_step = _iterator.next()).done) {
          context$1$0.next = 18;
          break;
        }

        cl = _step.value;
        context$1$0.next = 13;
        return this.dbSearch({
          subject: this.db.v('desc'),
          predicate: RDF + 'type',
          object: MESHV + cl
        }, {});

      case 13:
        result = context$1$0.sent;

        results = results.concat(result);

      case 15:
        _iteratorNormalCompletion = true;
        context$1$0.next = 9;
        break;

      case 18:
        context$1$0.next = 24;
        break;

      case 20:
        context$1$0.prev = 20;
        context$1$0.t0 = context$1$0['catch'](7);
        _didIteratorError = true;
        _iteratorError = context$1$0.t0;

      case 24:
        context$1$0.prev = 24;
        context$1$0.prev = 25;

        if (!_iteratorNormalCompletion && _iterator['return']) {
          _iterator['return']();
        }

      case 27:
        context$1$0.prev = 27;

        if (!_didIteratorError) {
          context$1$0.next = 30;
          break;
        }

        throw _iteratorError;

      case 30:
        return context$1$0.finish(27);

      case 31:
        return context$1$0.finish(24);

      case 32:
        return context$1$0.abrupt('return', _lodash2['default'].map(results, function (item) {
          return _this.formatID(item['desc'], format);
        }));

      case 33:
      case 'end':
        return context$1$0.stop();
    }
  }, callee$0$0, this, [[7, 20, 24, 32], [25,, 27, 31]]);
}));

/*
* Returns array of all chemical supplementary records
*/
MeshTree.prototype.getAllSCRChemicalUIs = _bluebird2['default'].coroutine(_regeneratorRuntime.mark(function callee$0$0(opts) {
  var format, result;
  return _regeneratorRuntime.wrap(function callee$0$0$(context$1$0) {
    var _this2 = this;

    while (1) switch (context$1$0.prev = context$1$0.next) {
      case 0:
        opts = opts || {};
        format = opts.format || 'rdf';
        context$1$0.next = 4;
        return this.dbSearch({
          subject: this.db.v('chemicalSCR'),
          predicate: RDF + 'type',
          object: MESHV + 'SCR_Chemical'
        }, {});

      case 4:
        result = context$1$0.sent;
        return context$1$0.abrupt('return', _lodash2['default'].map(result, function (item) {
          return _this2.formatID(item['chemicalSCR'], format);
        }));

      case 6:
      case 'end':
        return context$1$0.stop();
    }
  }, callee$0$0, this);
}));

/*
* Returns array of all disease (rare) supplementary records
*/
MeshTree.prototype.getAllSCRDiseaseUIs = _bluebird2['default'].coroutine(_regeneratorRuntime.mark(function callee$0$0(opts) {
  var format, result;
  return _regeneratorRuntime.wrap(function callee$0$0$(context$1$0) {
    var _this3 = this;

    while (1) switch (context$1$0.prev = context$1$0.next) {
      case 0:
        opts = opts || {};
        format = opts.format || 'rdf';
        context$1$0.next = 4;
        return this.dbSearch({
          subject: this.db.v('diseaseSCR'),
          predicate: RDF + 'type',
          object: MESHV + 'SCR_Disease'
        }, {});

      case 4:
        result = context$1$0.sent;
        return context$1$0.abrupt('return', _lodash2['default'].map(result, function (item) {
          return _this3.formatID(item['diseaseSCR'], format);
        }));

      case 6:
      case 'end':
        return context$1$0.stop();
    }
  }, callee$0$0, this);
}));

/*
* Returns array of all protocol (e.g., cancer-related) supplementary records
*/
MeshTree.prototype.getAllSCRProtocolUIs = _bluebird2['default'].coroutine(_regeneratorRuntime.mark(function callee$0$0(opts) {
  var format, result;
  return _regeneratorRuntime.wrap(function callee$0$0$(context$1$0) {
    var _this4 = this;

    while (1) switch (context$1$0.prev = context$1$0.next) {
      case 0:
        opts = opts || {};
        format = opts.format || 'rdf';
        context$1$0.next = 4;
        return this.dbSearch({
          subject: this.db.v('protocolSCR'),
          predicate: RDF + 'type',
          object: MESHV + 'SCR_Protocol'
        }, {});

      case 4:
        result = context$1$0.sent;
        return context$1$0.abrupt('return', _lodash2['default'].map(result, function (item) {
          return _this4.formatID(item['protocolSCR'], format);
        }));

      case 6:
      case 'end':
        return context$1$0.stop();
    }
  }, callee$0$0, this);
}));

/*
* Returns the cleaned text output of the wikipedia page corresponding to the descriptor record UI
*
* `level`:
*   `0` - abstract only
*   `1` - all text
*/
MeshTree.prototype.getWikiEntry = _bluebird2['default'].coroutine(_regeneratorRuntime.mark(function callee$0$0(opts) {
  var descUI, level, concept, wiki, text;
  return _regeneratorRuntime.wrap(function callee$0$0$(context$1$0) {
    while (1) switch (context$1$0.prev = context$1$0.next) {
      case 0:
        opts = opts || {};
        descUI = this.formatID(opts.id, 'mesh');
        level = opts.level || 0;
        context$1$0.next = 5;
        return this.getPrefTerm({ id: descUI });

      case 5:
        concept = context$1$0.sent;
        context$1$0.next = 8;
        return wikipedia.getMainSections(concept.replace(/ /g, '+'));

      case 8:
        wiki = context$1$0.sent;

        if (!(level === 0)) {
          context$1$0.next = 16;
          break;
        }

        text = '';

        _lodash2['default'].each(wiki, function (section) {
          if (section.sectionLevel === 0) {
            text += section.sectionText;
          }
        });

        // if no abstract, just return everything as if level = 1
        if (text.length === 0) {
          _lodash2['default'].each(wiki, function (section) {
            text += section.sectionText;
          });
        }

        return context$1$0.abrupt('return', text);

      case 16:
        text = '';

        _lodash2['default'].each(wiki, function (section) {
          text += section.sectionText;
        });

        return context$1$0.abrupt('return', text);

      case 19:
      case 'end':
        return context$1$0.stop();
    }
  }, callee$0$0, this);
}));

/*
* Returns array of tree numbers for descriptor record unique identifier
*
* Example: 'D000001' returns ['D03.438.221.173']
*/
MeshTree.prototype.getTreeNumbers = _bluebird2['default'].coroutine(_regeneratorRuntime.mark(function callee$0$0(opts) {
  var descUI, format, result, treeNumbers;
  return _regeneratorRuntime.wrap(function callee$0$0$(context$1$0) {
    var _this5 = this;

    while (1) switch (context$1$0.prev = context$1$0.next) {
      case 0:
        opts = opts || {};
        descUI = this.formatID(opts.id, 'mesh');
        format = opts.format || 'rdf';
        context$1$0.next = 5;
        return this.dbSearch({
          subject: MESH + descUI,
          predicate: MESHV + 'treeNumber',
          object: this.db.v('treeNumber')
        }, {});

      case 5:
        result = context$1$0.sent;
        treeNumbers = _lodash2['default'].map(result, function (item) {
          return _this5.formatID(item['treeNumber'], format);
        });
        return context$1$0.abrupt('return', treeNumbers);

      case 8:
      case 'end':
        return context$1$0.stop();
    }
  }, callee$0$0, this);
}));

/*
* Returns descriptor record unique identifier by tree number
*
* Example: 'D03.438.221.173' returns 'D000001'
*/
MeshTree.prototype.treeNumberToUI = _bluebird2['default'].coroutine(_regeneratorRuntime.mark(function callee$0$0(opts) {
  var treeNum, format, result;
  return _regeneratorRuntime.wrap(function callee$0$0$(context$1$0) {
    while (1) switch (context$1$0.prev = context$1$0.next) {
      case 0:
        opts = opts || {};
        treeNum = this.formatID(opts.treeNum, 'mesh');
        format = opts.format || 'rdf';
        context$1$0.next = 5;
        return this.dbSearch({
          subject: this.db.v('descUI'),
          predicate: MESHV + 'treeNumber',
          object: MESH + treeNum
        }, {});

      case 5:
        result = context$1$0.sent;

        if (!_lodash2['default'].isEmpty(result)) {
          context$1$0.next = 10;
          break;
        }

        return context$1$0.abrupt('return', null);

      case 10:
        return context$1$0.abrupt('return', this.formatID(result[0]['descUI'], format));

      case 11:
      case 'end':
        return context$1$0.stop();
    }
  }, callee$0$0, this);
}));

/*
* Returns top-level MeSH category
*
* Example: 'D000001' returns 'Chemicals and Drugs'
*/
MeshTree.prototype.getCategory = _bluebird2['default'].coroutine(_regeneratorRuntime.mark(function callee$0$0(opts) {
  var descUI, categoryMap, treeNums, categories;
  return _regeneratorRuntime.wrap(function callee$0$0$(context$1$0) {
    while (1) switch (context$1$0.prev = context$1$0.next) {
      case 0:
        opts = opts || {};
        descUI = this.formatID(opts.id, 'mesh');
        categoryMap = new _Map([['A', 'Anatomy'], ['B', 'Organisms'], ['C', 'Diseases'], ['D', 'Chemicals and Drugs'], ['E', 'Analytical, Diagnostic and Therapeutic Techniques and Equipment'], ['F', 'Psychiatry and Psychology'], ['G', 'Phenomena and Processes'], ['H', 'Disciplines and Occupations'], ['I', 'Anthropology, Education, Sociology and Social Phenomena'], ['J', 'Technology, Industry, Agriculture'], ['K', 'Humanities'], ['L', 'Information Science'], ['M', 'Named Groups'], ['N', 'Health Care'], ['V', 'Publication Characteristics'], ['Z', 'Geographicals']]);
        context$1$0.next = 5;
        return this.getTreeNumbers({ id: descUI, format: 'mesh' });

      case 5:
        treeNums = context$1$0.sent;
        categories = treeNums.map(function (t) {
          return t[0];
        });
        return context$1$0.abrupt('return', categoryMap.get((0, _lodash2['default'])(categories).countBy().pairs().max(function (kv) {
          return kv[1];
        })[0]));

      case 8:
      case 'end':
        return context$1$0.stop();
    }
  }, callee$0$0, this);
}));

/*
* Returns the record preferred term by descriptor record unique identifier
* (i.e., the preferred term of the preferred concept)
*
* Example: 'D000001' returns 'Calcimycin'
*/
MeshTree.prototype.getPrefTerm = _bluebird2['default'].coroutine(_regeneratorRuntime.mark(function callee$0$0(opts) {
  var descUI, result1, result2, result2_en;
  return _regeneratorRuntime.wrap(function callee$0$0$(context$1$0) {
    while (1) switch (context$1$0.prev = context$1$0.next) {
      case 0:
        opts = opts || {};
        descUI = this.formatID(opts.id, 'mesh');
        context$1$0.next = 4;
        return this.dbSearch({
          subject: MESH + descUI,
          predicate: MESHV + 'preferredTerm',
          object: this.db.v('preferredTermUI')
        }, {});

      case 4:
        result1 = context$1$0.sent;
        context$1$0.next = 7;
        return this.dbSearch({
          subject: result1[0]['preferredTermUI'],
          predicate: MESHV + 'prefLabel',
          object: this.db.v('term')
        }, {});

      case 7:
        result2 = context$1$0.sent;

        if (!_lodash2['default'].isEmpty(result2)) {
          context$1$0.next = 10;
          break;
        }

        return context$1$0.abrupt('return', null);

      case 10:
        result2_en = result2.filter(function (r) {
          return r['term'].endsWith('@en');
        });
        return context$1$0.abrupt('return', result2_en[0]['term'].replace(/\"/g, '').replace(/@en/, ''));

      case 12:
      case 'end':
        return context$1$0.stop();
    }
  }, callee$0$0, this);
}));

/*
* Returns all terms by descriptor record unique identifier
* (i.e., all terms for all concepts, both preferred and not)
*
* Example: 'D000001' returns [ 'A23187, Antibiotic', 'Antibiotic A23187', 'A23187', 'A 23187', 'A-23187', 'Calcimycin' ]
*/
MeshTree.prototype.getAllTerms = _bluebird2['default'].coroutine(_regeneratorRuntime.mark(function callee$0$0(opts) {
  var descUI, allTermLabels, _arr, _i, pred1, conceptUIs, _iteratorNormalCompletion2, _didIteratorError2, _iteratorError2, _iterator2, _step2, item, conceptUI, _arr2, _i2, pred2, termUIs, _iteratorNormalCompletion3, _didIteratorError3, _iteratorError3, _iterator3, _step3, _item, termUI, _arr3, _i3, pred3, terms, terms_en;

  return _regeneratorRuntime.wrap(function callee$0$0$(context$1$0) {
    while (1) switch (context$1$0.prev = context$1$0.next) {
      case 0:
        opts = opts || {};
        descUI = this.formatID(opts.id, 'mesh');
        allTermLabels = [];
        _arr = ['concept', 'preferredConcept'];
        _i = 0;

      case 5:
        if (!(_i < _arr.length)) {
          context$1$0.next = 85;
          break;
        }

        pred1 = _arr[_i];
        context$1$0.next = 9;
        return this.dbSearch({
          subject: MESH + descUI,
          predicate: MESHV + pred1,
          object: this.db.v('conceptUI')
        }, {});

      case 9:
        conceptUIs = context$1$0.sent;
        _iteratorNormalCompletion2 = true;
        _didIteratorError2 = false;
        _iteratorError2 = undefined;
        context$1$0.prev = 13;
        _iterator2 = _getIterator(conceptUIs);

      case 15:
        if (_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done) {
          context$1$0.next = 68;
          break;
        }

        item = _step2.value;
        conceptUI = item['conceptUI'];
        _arr2 = ['term', 'preferredTerm'];
        _i2 = 0;

      case 20:
        if (!(_i2 < _arr2.length)) {
          context$1$0.next = 65;
          break;
        }

        pred2 = _arr2[_i2];
        context$1$0.next = 24;
        return this.dbSearch({
          subject: conceptUI,
          predicate: MESHV + pred2,
          object: this.db.v('termUI')
        }, {});

      case 24:
        termUIs = context$1$0.sent;
        _iteratorNormalCompletion3 = true;
        _didIteratorError3 = false;
        _iteratorError3 = undefined;
        context$1$0.prev = 28;
        _iterator3 = _getIterator(termUIs);

      case 30:
        if (_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done) {
          context$1$0.next = 48;
          break;
        }

        _item = _step3.value;
        termUI = _item['termUI'];
        _arr3 = ['label', 'altLabel', 'prefLabel'];
        _i3 = 0;

      case 35:
        if (!(_i3 < _arr3.length)) {
          context$1$0.next = 45;
          break;
        }

        pred3 = _arr3[_i3];
        context$1$0.next = 39;
        return this.dbSearch({
          subject: termUI,
          predicate: MESHV + pred3,
          object: this.db.v('label')
        }, {});

      case 39:
        terms = context$1$0.sent;
        terms_en = terms.filter(function (r) {
          return r['label'].endsWith('@en');
        });

        terms_en.forEach(function (term) {
          return allTermLabels.push(term['label'].replace(/\"/g, '').replace(/@en/, ''));
        });

      case 42:
        _i3++;
        context$1$0.next = 35;
        break;

      case 45:
        _iteratorNormalCompletion3 = true;
        context$1$0.next = 30;
        break;

      case 48:
        context$1$0.next = 54;
        break;

      case 50:
        context$1$0.prev = 50;
        context$1$0.t0 = context$1$0['catch'](28);
        _didIteratorError3 = true;
        _iteratorError3 = context$1$0.t0;

      case 54:
        context$1$0.prev = 54;
        context$1$0.prev = 55;

        if (!_iteratorNormalCompletion3 && _iterator3['return']) {
          _iterator3['return']();
        }

      case 57:
        context$1$0.prev = 57;

        if (!_didIteratorError3) {
          context$1$0.next = 60;
          break;
        }

        throw _iteratorError3;

      case 60:
        return context$1$0.finish(57);

      case 61:
        return context$1$0.finish(54);

      case 62:
        _i2++;
        context$1$0.next = 20;
        break;

      case 65:
        _iteratorNormalCompletion2 = true;
        context$1$0.next = 15;
        break;

      case 68:
        context$1$0.next = 74;
        break;

      case 70:
        context$1$0.prev = 70;
        context$1$0.t1 = context$1$0['catch'](13);
        _didIteratorError2 = true;
        _iteratorError2 = context$1$0.t1;

      case 74:
        context$1$0.prev = 74;
        context$1$0.prev = 75;

        if (!_iteratorNormalCompletion2 && _iterator2['return']) {
          _iterator2['return']();
        }

      case 77:
        context$1$0.prev = 77;

        if (!_didIteratorError2) {
          context$1$0.next = 80;
          break;
        }

        throw _iteratorError2;

      case 80:
        return context$1$0.finish(77);

      case 81:
        return context$1$0.finish(74);

      case 82:
        _i++;
        context$1$0.next = 5;
        break;

      case 85:
        return context$1$0.abrupt('return', allTermLabels);

      case 86:
      case 'end':
        return context$1$0.stop();
    }
  }, callee$0$0, this, [[13, 70, 74, 82], [28, 50, 54, 62], [55,, 57, 61], [75,, 77, 81]]);
}));

/*
* Returns scope note for descriptor record unique identifier
* (scope notes are contained in the preferred concept record)
*
* Example: 'D000001', via concept 'M0000001', returns 'An ionophorous, polyether antibiotic from Streptomyces chartreusensis. It binds and transports CALCIUM and other divalent cations across membranes and uncouples oxidative phosphorylation while inhibiting ATPase of rat liver mitochondria. The substance is used mostly as a biochemical tool to study the role of divalent cations in various biological systems.'
*/
MeshTree.prototype.getScopeNote = _bluebird2['default'].coroutine(_regeneratorRuntime.mark(function callee$0$0(opts) {
  var descUI, prefConcept, result, result_en;
  return _regeneratorRuntime.wrap(function callee$0$0$(context$1$0) {
    while (1) switch (context$1$0.prev = context$1$0.next) {
      case 0:
        opts = opts || {};
        descUI = this.formatID(opts.id, 'mesh');
        context$1$0.next = 4;
        return this.dbSearch({
          subject: MESH + descUI,
          predicate: MESHV + 'preferredConcept',
          object: this.db.v('conceptUI')
        }, {});

      case 4:
        prefConcept = context$1$0.sent;

        if (!_lodash2['default'].isEmpty(prefConcept)) {
          context$1$0.next = 7;
          break;
        }

        return context$1$0.abrupt('return', '');

      case 7:
        context$1$0.next = 9;
        return this.dbSearch({
          subject: prefConcept[0]['conceptUI'],
          predicate: MESHV + 'scopeNote',
          object: this.db.v('scopeNote')
        }, {});

      case 9:
        result = context$1$0.sent;

        if (!_lodash2['default'].isEmpty(result)) {
          context$1$0.next = 14;
          break;
        }

        return context$1$0.abrupt('return', '');

      case 14:
        result_en = result.filter(function (r) {
          return r['scopeNote'].endsWith('@en');
        });
        return context$1$0.abrupt('return', result_en[0]['scopeNote'].replace(/\"/g, '').replace(/@en/, ''));

      case 16:
      case 'end':
        return context$1$0.stop();
    }
  }, callee$0$0, this);
}));

/*
* Returns parent descriptor records UIs
* (returns an array as records can exist in multiple tree branches)
*
* Example: 'D000001' returns ['D001583']
*          'D005138' returns ['D006197', 'D005123']
*
* If id provided is SCR,
* returns parent descriptor record UIs mapped from supplementary concept record UI
*
* Example: 'C025735' returns ['D001286', 'D002164', 'D012602']
*/
MeshTree.prototype.getParents = _bluebird2['default'].coroutine(_regeneratorRuntime.mark(function callee$0$0(opts) {
  var ui, format, result;
  return _regeneratorRuntime.wrap(function callee$0$0$(context$1$0) {
    var _this6 = this;

    while (1) switch (context$1$0.prev = context$1$0.next) {
      case 0:
        opts = opts || {};
        ui = this.formatID(opts.id, 'mesh');
        format = opts.format || 'rdf';

        if (!/^D\d+$/.test(ui)) {
          context$1$0.next = 10;
          break;
        }

        context$1$0.next = 6;
        return this.dbSearch({
          subject: MESH + ui,
          predicate: MESHV + 'broaderDescriptor',
          object: this.db.v('descUI')
        }, {});

      case 6:
        result = context$1$0.sent;
        return context$1$0.abrupt('return', _lodash2['default'].unique(result.map(function (res) {
          return _this6.formatID(res['descUI'], format);
        })));

      case 10:
        if (!/^C\d+$/.test(ui)) {
          context$1$0.next = 17;
          break;
        }

        context$1$0.next = 13;
        return this.dbSearch({
          subject: MESH + ui,
          predicate: MESHV + 'preferredMappedTo',
          object: this.db.v('descUI')
        }, {});

      case 13:
        result = context$1$0.sent;
        return context$1$0.abrupt('return', _lodash2['default'].unique(result.map(function (res) {
          return _this6.formatID(res['descUI'], format).replace(/Q\d+/g, '');
        })));

      case 17:
        return context$1$0.abrupt('return', []);

      case 18:
      case 'end':
        return context$1$0.stop();
    }
  }, callee$0$0, this);
}));

/*
* Returns ancestor descriptor records UIs
* (returns an array)
*
* Example: 'D000001' returns ['D001583', 'D006574', 'D006571']
*          'D005138' returns ['D005123', 'D006197', 'D005145', 'D012679', 'D034582', 'D006257', 'D001829']
*/
MeshTree.prototype.getAncestors = _bluebird2['default'].coroutine(_regeneratorRuntime.mark(function callee$0$0(opts) {
  var descUI, format, parents, nodes, hasParents, _loop, _iteratorNormalCompletion4, _didIteratorError4, _iteratorError4, _iterator4, _step4;

  return _regeneratorRuntime.wrap(function callee$0$0$(context$1$0) {
    var _this7 = this;

    while (1) switch (context$1$0.prev = context$1$0.next) {
      case 0:
        opts = opts || {};
        descUI = this.formatID(opts.id, 'mesh');
        format = opts.format || 'rdf';
        parents = [];
        nodes = [descUI];
        hasParents = true;
        _loop = _regeneratorRuntime.mark(function callee$1$0() {
          var nodesTemp, node, parentsTemp;
          return _regeneratorRuntime.wrap(function callee$1$0$(context$2$0) {
            while (1) switch (context$2$0.prev = context$2$0.next) {
              case 0:
                nodesTemp = [];
                _iteratorNormalCompletion4 = true;
                _didIteratorError4 = false;
                _iteratorError4 = undefined;
                context$2$0.prev = 4;
                _iterator4 = _getIterator(nodes);

              case 6:
                if (_iteratorNormalCompletion4 = (_step4 = _iterator4.next()).done) {
                  context$2$0.next = 16;
                  break;
                }

                node = _step4.value;
                context$2$0.next = 10;
                return this.getParents({ id: node, format: 'mesh' });

              case 10:
                parentsTemp = context$2$0.sent;

                _lodash2['default'].each(parentsTemp, function (p) {
                  return nodesTemp.push(p);
                });
                _lodash2['default'].each(parentsTemp, function (p) {
                  return parents.push(p);
                });

              case 13:
                _iteratorNormalCompletion4 = true;
                context$2$0.next = 6;
                break;

              case 16:
                context$2$0.next = 22;
                break;

              case 18:
                context$2$0.prev = 18;
                context$2$0.t0 = context$2$0['catch'](4);
                _didIteratorError4 = true;
                _iteratorError4 = context$2$0.t0;

              case 22:
                context$2$0.prev = 22;
                context$2$0.prev = 23;

                if (!_iteratorNormalCompletion4 && _iterator4['return']) {
                  _iterator4['return']();
                }

              case 25:
                context$2$0.prev = 25;

                if (!_didIteratorError4) {
                  context$2$0.next = 28;
                  break;
                }

                throw _iteratorError4;

              case 28:
                return context$2$0.finish(25);

              case 29:
                return context$2$0.finish(22);

              case 30:

                nodes = nodesTemp;
                hasParents = nodesTemp.length > 0;

              case 32:
              case 'end':
                return context$2$0.stop();
            }
          }, callee$1$0, _this7, [[4, 18, 22, 30], [23,, 25, 29]]);
        });

      case 7:
        if (!hasParents) {
          context$1$0.next = 11;
          break;
        }

        return context$1$0.delegateYield(_loop(), 't0', 9);

      case 9:
        context$1$0.next = 7;
        break;

      case 11:
        return context$1$0.abrupt('return', _lodash2['default'].unique(parents).map(function (p) {
          return _this7.formatID(p, format);
        }));

      case 12:
      case 'end':
        return context$1$0.stop();
    }
  }, callee$0$0, this);
}));

/*
* Returns children descriptor records UIs
* (immediate, not descendants)
*
* Example: 'D012343' returns ['D012345', 'D000926', 'D012346']
*/
MeshTree.prototype.getChildren = _bluebird2['default'].coroutine(_regeneratorRuntime.mark(function callee$0$0(opts) {
  var descUI, format, childrenDescUIs, treeNums, _iteratorNormalCompletion5, _didIteratorError5, _iteratorError5, _iterator5, _step5, treeNum, result, _iteratorNormalCompletion6, _didIteratorError6, _iteratorError6, _iterator6, _step6, res, _treeNum, ui;

  return _regeneratorRuntime.wrap(function callee$0$0$(context$1$0) {
    var _this8 = this;

    while (1) switch (context$1$0.prev = context$1$0.next) {
      case 0:
        opts = opts || {};
        descUI = this.formatID(opts.id, 'mesh');
        format = opts.format || 'rdf';
        childrenDescUIs = [];
        context$1$0.next = 6;
        return this.getTreeNumbers({ id: descUI, format: 'mesh' });

      case 6:
        treeNums = context$1$0.sent;
        _iteratorNormalCompletion5 = true;
        _didIteratorError5 = false;
        _iteratorError5 = undefined;
        context$1$0.prev = 10;
        _iterator5 = _getIterator(treeNums);

      case 12:
        if (_iteratorNormalCompletion5 = (_step5 = _iterator5.next()).done) {
          context$1$0.next = 50;
          break;
        }

        treeNum = _step5.value;
        context$1$0.next = 16;
        return this.dbSearch({
          subject: this.db.v('treeNum'),
          predicate: MESHV + 'parentTreeNumber',
          object: MESH + treeNum
        }, {});

      case 16:
        result = context$1$0.sent;

        if (_lodash2['default'].isEmpty(result)) {
          context$1$0.next = 47;
          break;
        }

        _iteratorNormalCompletion6 = true;
        _didIteratorError6 = false;
        _iteratorError6 = undefined;
        context$1$0.prev = 21;
        _iterator6 = _getIterator(result);

      case 23:
        if (_iteratorNormalCompletion6 = (_step6 = _iterator6.next()).done) {
          context$1$0.next = 33;
          break;
        }

        res = _step6.value;
        _treeNum = this.formatID(res['treeNum'], 'mesh');
        context$1$0.next = 28;
        return this.treeNumberToUI({ treeNum: _treeNum, format: 'mesh' });

      case 28:
        ui = context$1$0.sent;

        childrenDescUIs.push(ui);

      case 30:
        _iteratorNormalCompletion6 = true;
        context$1$0.next = 23;
        break;

      case 33:
        context$1$0.next = 39;
        break;

      case 35:
        context$1$0.prev = 35;
        context$1$0.t0 = context$1$0['catch'](21);
        _didIteratorError6 = true;
        _iteratorError6 = context$1$0.t0;

      case 39:
        context$1$0.prev = 39;
        context$1$0.prev = 40;

        if (!_iteratorNormalCompletion6 && _iterator6['return']) {
          _iterator6['return']();
        }

      case 42:
        context$1$0.prev = 42;

        if (!_didIteratorError6) {
          context$1$0.next = 45;
          break;
        }

        throw _iteratorError6;

      case 45:
        return context$1$0.finish(42);

      case 46:
        return context$1$0.finish(39);

      case 47:
        _iteratorNormalCompletion5 = true;
        context$1$0.next = 12;
        break;

      case 50:
        context$1$0.next = 56;
        break;

      case 52:
        context$1$0.prev = 52;
        context$1$0.t1 = context$1$0['catch'](10);
        _didIteratorError5 = true;
        _iteratorError5 = context$1$0.t1;

      case 56:
        context$1$0.prev = 56;
        context$1$0.prev = 57;

        if (!_iteratorNormalCompletion5 && _iterator5['return']) {
          _iterator5['return']();
        }

      case 59:
        context$1$0.prev = 59;

        if (!_didIteratorError5) {
          context$1$0.next = 62;
          break;
        }

        throw _iteratorError5;

      case 62:
        return context$1$0.finish(59);

      case 63:
        return context$1$0.finish(56);

      case 64:
        return context$1$0.abrupt('return', childrenDescUIs.map(function (ui) {
          return _this8.formatID(ui, format);
        }));

      case 65:
      case 'end':
        return context$1$0.stop();
    }
  }, callee$0$0, this, [[10, 52, 56, 64], [21, 35, 39, 47], [40,, 42, 46], [57,, 59, 63]]);
}));

/*
* Returns sibling descriptor records UIs
* (across all branches a descriptor record may exist under)
*
* Example: 'D015834' returns ['D012345', 'D000926', 'D012346']
*/
MeshTree.prototype.getSiblings = _bluebird2['default'].coroutine(_regeneratorRuntime.mark(function callee$0$0(opts) {
  var descUI, format, siblingDescUIs, parentDescUIs, _iteratorNormalCompletion7, _didIteratorError7, _iteratorError7, _iterator7, _step7, _parent, childrenDescUIs;

  return _regeneratorRuntime.wrap(function callee$0$0$(context$1$0) {
    var _this9 = this;

    while (1) switch (context$1$0.prev = context$1$0.next) {
      case 0:
        opts = opts || {};
        descUI = this.formatID(opts.id, 'mesh');
        format = opts.format || 'rdf';
        siblingDescUIs = [];
        context$1$0.next = 6;
        return this.getParents({ id: descUI, format: 'mesh' });

      case 6:
        parentDescUIs = context$1$0.sent;
        _iteratorNormalCompletion7 = true;
        _didIteratorError7 = false;
        _iteratorError7 = undefined;
        context$1$0.prev = 10;
        _iterator7 = _getIterator(parentDescUIs);

      case 12:
        if (_iteratorNormalCompletion7 = (_step7 = _iterator7.next()).done) {
          context$1$0.next = 21;
          break;
        }

        _parent = _step7.value;
        context$1$0.next = 16;
        return this.getChildren({ id: _parent, format: 'mesh' });

      case 16:
        childrenDescUIs = context$1$0.sent;

        _lodash2['default'].forEach(childrenDescUIs, function (ui) {
          return siblingDescUIs.push(ui);
        });

      case 18:
        _iteratorNormalCompletion7 = true;
        context$1$0.next = 12;
        break;

      case 21:
        context$1$0.next = 27;
        break;

      case 23:
        context$1$0.prev = 23;
        context$1$0.t0 = context$1$0['catch'](10);
        _didIteratorError7 = true;
        _iteratorError7 = context$1$0.t0;

      case 27:
        context$1$0.prev = 27;
        context$1$0.prev = 28;

        if (!_iteratorNormalCompletion7 && _iterator7['return']) {
          _iterator7['return']();
        }

      case 30:
        context$1$0.prev = 30;

        if (!_didIteratorError7) {
          context$1$0.next = 33;
          break;
        }

        throw _iteratorError7;

      case 33:
        return context$1$0.finish(30);

      case 34:
        return context$1$0.finish(27);

      case 35:

        _lodash2['default'].remove(siblingDescUIs, function (ui) {
          return ui === descUI;
        });

        return context$1$0.abrupt('return', siblingDescUIs.map(function (ui) {
          return _this9.formatID(ui, format);
        }));

      case 37:
      case 'end':
        return context$1$0.stop();
    }
  }, callee$0$0, this, [[10, 23, 27, 35], [28,, 30, 34]]);
}));

/*
* Returns descriptor records UI of closest common ancestors of two or more descriptor record UIs
* (if a descriptor exists in more than one place on the tree, there will be more than one common ancestor)
*
* Example: ['D000926', 'D012345'] returns ['D012343']
*/
MeshTree.prototype.getCommonAncestors = _bluebird2['default'].coroutine(_regeneratorRuntime.mark(function callee$0$0(opts) {
  var descUIArray, format, commonAncestorsDescUIs, commonAncestorsTreeNums, treeNumsArray, _iteratorNormalCompletion8, _didIteratorError8, _iteratorError8, _iterator8, _step8, descUI, treeNums, treeNumPermutations, _iteratorNormalCompletion9, _didIteratorError9, _iteratorError9, _loop2, _iterator9, _step9, _iteratorNormalCompletion10, _didIteratorError10, _iteratorError10, _iterator10, _step10, _iteratorNormalCompletion11, _didIteratorError11, _iteratorError11, _iterator11, _step11;

  return _regeneratorRuntime.wrap(function callee$0$0$(context$1$0) {
    var _this10 = this;

    while (1) switch (context$1$0.prev = context$1$0.next) {
      case 0:
        opts = opts || {};

        if (_lodash2['default'].isArray(opts.ids)) {
          context$1$0.next = 3;
          break;
        }

        throw new Error('opts.ids not an array.');

      case 3:
        descUIArray = opts.ids.map(function (id) {
          return _this10.formatID(id, 'mesh');
        });
        format = opts.format || 'rdf';
        commonAncestorsDescUIs = [];
        commonAncestorsTreeNums = [];
        treeNumsArray = [];
        _iteratorNormalCompletion8 = true;
        _didIteratorError8 = false;
        _iteratorError8 = undefined;
        context$1$0.prev = 11;
        _iterator8 = _getIterator(descUIArray);

      case 13:
        if (_iteratorNormalCompletion8 = (_step8 = _iterator8.next()).done) {
          context$1$0.next = 22;
          break;
        }

        descUI = _step8.value;
        context$1$0.next = 17;
        return this.getTreeNumbers({ id: descUI, format: 'mesh' });

      case 17:
        treeNums = context$1$0.sent;

        treeNumsArray.push(treeNums);

      case 19:
        _iteratorNormalCompletion8 = true;
        context$1$0.next = 13;
        break;

      case 22:
        context$1$0.next = 28;
        break;

      case 24:
        context$1$0.prev = 24;
        context$1$0.t0 = context$1$0['catch'](11);
        _didIteratorError8 = true;
        _iteratorError8 = context$1$0.t0;

      case 28:
        context$1$0.prev = 28;
        context$1$0.prev = 29;

        if (!_iteratorNormalCompletion8 && _iterator8['return']) {
          _iterator8['return']();
        }

      case 31:
        context$1$0.prev = 31;

        if (!_didIteratorError8) {
          context$1$0.next = 34;
          break;
        }

        throw _iteratorError8;

      case 34:
        return context$1$0.finish(31);

      case 35:
        return context$1$0.finish(28);

      case 36:
        treeNumPermutations = _helperPermutations2['default'].apply(null, treeNumsArray);
        _iteratorNormalCompletion9 = true;
        _didIteratorError9 = false;
        _iteratorError9 = undefined;
        context$1$0.prev = 40;
        _loop2 = _regeneratorRuntime.mark(function callee$1$0() {
          var permut, depth, branches, _loop3, _ret3, commonAncestorTreeNum, isBroader, indices, idx, commonAncestorDescUI;

          return _regeneratorRuntime.wrap(function callee$1$0$(context$2$0) {
            while (1) switch (context$2$0.prev = context$2$0.next) {
              case 0:
                permut = _step9.value;
                depth = 0;
                branches = [];
                _iteratorNormalCompletion10 = true;
                _didIteratorError10 = false;
                _iteratorError10 = undefined;
                context$2$0.prev = 6;

                _loop3 = function () {
                  var branch = _step10.value;

                  var isCommonBranch = _lodash2['default'].all(_lodash2['default'].map(permut.slice(1, permut.length), function (x) {
                    return branch === x.split('.')[depth];
                  }));

                  if (isCommonBranch) {
                    branches.push(branch);
                    ++depth;
                  } else {
                    return 'break';
                  }
                };

                _iterator10 = _getIterator(permut[0].split('.'));

              case 9:
                if (_iteratorNormalCompletion10 = (_step10 = _iterator10.next()).done) {
                  context$2$0.next = 16;
                  break;
                }

                _ret3 = _loop3();

                if (!(_ret3 === 'break')) {
                  context$2$0.next = 13;
                  break;
                }

                return context$2$0.abrupt('break', 16);

              case 13:
                _iteratorNormalCompletion10 = true;
                context$2$0.next = 9;
                break;

              case 16:
                context$2$0.next = 22;
                break;

              case 18:
                context$2$0.prev = 18;
                context$2$0.t0 = context$2$0['catch'](6);
                _didIteratorError10 = true;
                _iteratorError10 = context$2$0.t0;

              case 22:
                context$2$0.prev = 22;
                context$2$0.prev = 23;

                if (!_iteratorNormalCompletion10 && _iterator10['return']) {
                  _iterator10['return']();
                }

              case 25:
                context$2$0.prev = 25;

                if (!_didIteratorError10) {
                  context$2$0.next = 28;
                  break;
                }

                throw _iteratorError10;

              case 28:
                return context$2$0.finish(25);

              case 29:
                return context$2$0.finish(22);

              case 30:
                commonAncestorTreeNum = branches.join('.');
                isBroader = _lodash2['default'].any(_lodash2['default'].map(commonAncestorsTreeNums, function (t) {
                  return _lodash2['default'].startsWith(t, commonAncestorTreeNum);
                }));

                if (!(commonAncestorTreeNum && !isBroader)) {
                  context$2$0.next = 59;
                  break;
                }

                indices = [];

                for (idx = 0; idx < commonAncestorsTreeNums.length; idx++) {
                  if (_lodash2['default'].startsWith(commonAncestorTreeNum, commonAncestorsTreeNums[idx])) {
                    indices.push(idx);
                  }
                }
                _iteratorNormalCompletion11 = true;
                _didIteratorError11 = false;
                _iteratorError11 = undefined;
                context$2$0.prev = 38;
                for (_iterator11 = _getIterator(indices); !(_iteratorNormalCompletion11 = (_step11 = _iterator11.next()).done); _iteratorNormalCompletion11 = true) {
                  idx = _step11.value;

                  commonAncestorsDescUIs.splice(idx, 1);
                  commonAncestorsTreeNums.splice(idx, 1);
                }

                // add common ancestor to array
                context$2$0.next = 46;
                break;

              case 42:
                context$2$0.prev = 42;
                context$2$0.t1 = context$2$0['catch'](38);
                _didIteratorError11 = true;
                _iteratorError11 = context$2$0.t1;

              case 46:
                context$2$0.prev = 46;
                context$2$0.prev = 47;

                if (!_iteratorNormalCompletion11 && _iterator11['return']) {
                  _iterator11['return']();
                }

              case 49:
                context$2$0.prev = 49;

                if (!_didIteratorError11) {
                  context$2$0.next = 52;
                  break;
                }

                throw _iteratorError11;

              case 52:
                return context$2$0.finish(49);

              case 53:
                return context$2$0.finish(46);

              case 54:
                context$2$0.next = 56;
                return this.treeNumberToUI({ treeNum: commonAncestorTreeNum, format: 'mesh' });

              case 56:
                commonAncestorDescUI = context$2$0.sent;

                commonAncestorsDescUIs.push(commonAncestorDescUI);
                commonAncestorsTreeNums.push(commonAncestorTreeNum);

              case 59:
              case 'end':
                return context$2$0.stop();
            }
          }, callee$1$0, _this10, [[6, 18, 22, 30], [23,, 25, 29], [38, 42, 46, 54], [47,, 49, 53]]);
        });
        _iterator9 = _getIterator(treeNumPermutations);

      case 43:
        if (_iteratorNormalCompletion9 = (_step9 = _iterator9.next()).done) {
          context$1$0.next = 48;
          break;
        }

        return context$1$0.delegateYield(_loop2(), 't1', 45);

      case 45:
        _iteratorNormalCompletion9 = true;
        context$1$0.next = 43;
        break;

      case 48:
        context$1$0.next = 54;
        break;

      case 50:
        context$1$0.prev = 50;
        context$1$0.t2 = context$1$0['catch'](40);
        _didIteratorError9 = true;
        _iteratorError9 = context$1$0.t2;

      case 54:
        context$1$0.prev = 54;
        context$1$0.prev = 55;

        if (!_iteratorNormalCompletion9 && _iterator9['return']) {
          _iterator9['return']();
        }

      case 57:
        context$1$0.prev = 57;

        if (!_didIteratorError9) {
          context$1$0.next = 60;
          break;
        }

        throw _iteratorError9;

      case 60:
        return context$1$0.finish(57);

      case 61:
        return context$1$0.finish(54);

      case 62:
        return context$1$0.abrupt('return', commonAncestorsDescUIs.map(function (ui) {
          return _this10.formatID(ui, format);
        }));

      case 63:
      case 'end':
        return context$1$0.stop();
    }
  }, callee$0$0, this, [[11, 24, 28, 36], [29,, 31, 35], [40, 50, 54, 62], [55,, 57, 61]]);
}));

/*
* Tests whether or not id2 is a descendant of id1 (child of >=1 depth)
*/
MeshTree.prototype.isDescendantOf = _bluebird2['default'].coroutine(_regeneratorRuntime.mark(function callee$0$0(id1, id2) {
  var descUI1, descUI2, parentsAll, nodes, hasParents, _loop4, _iteratorNormalCompletion12, _didIteratorError12, _iteratorError12, _iterator12, _step12;

  return _regeneratorRuntime.wrap(function callee$0$0$(context$1$0) {
    var _this11 = this;

    while (1) switch (context$1$0.prev = context$1$0.next) {
      case 0:
        descUI1 = this.formatID(id1, 'mesh');
        descUI2 = this.formatID(id2, 'mesh');

        if (!(descUI1 === descUI2)) {
          context$1$0.next = 4;
          break;
        }

        return context$1$0.abrupt('return', false);

      case 4:
        parentsAll = [];
        nodes = [descUI2];
        hasParents = true;
        _loop4 = _regeneratorRuntime.mark(function callee$1$0() {
          var nodesTemp, node, parentsTemp;
          return _regeneratorRuntime.wrap(function callee$1$0$(context$2$0) {
            while (1) switch (context$2$0.prev = context$2$0.next) {
              case 0:
                nodesTemp = [];
                _iteratorNormalCompletion12 = true;
                _didIteratorError12 = false;
                _iteratorError12 = undefined;
                context$2$0.prev = 4;
                _iterator12 = _getIterator(nodes);

              case 6:
                if (_iteratorNormalCompletion12 = (_step12 = _iterator12.next()).done) {
                  context$2$0.next = 15;
                  break;
                }

                node = _step12.value;
                context$2$0.next = 10;
                return this.getParents({ id: node, format: 'mesh' });

              case 10:
                parentsTemp = context$2$0.sent;

                _lodash2['default'].each(parentsTemp, function (p) {
                  parentsAll.push(p);
                  nodesTemp.push(p);
                });

              case 12:
                _iteratorNormalCompletion12 = true;
                context$2$0.next = 6;
                break;

              case 15:
                context$2$0.next = 21;
                break;

              case 17:
                context$2$0.prev = 17;
                context$2$0.t0 = context$2$0['catch'](4);
                _didIteratorError12 = true;
                _iteratorError12 = context$2$0.t0;

              case 21:
                context$2$0.prev = 21;
                context$2$0.prev = 22;

                if (!_iteratorNormalCompletion12 && _iterator12['return']) {
                  _iterator12['return']();
                }

              case 24:
                context$2$0.prev = 24;

                if (!_didIteratorError12) {
                  context$2$0.next = 27;
                  break;
                }

                throw _iteratorError12;

              case 27:
                return context$2$0.finish(24);

              case 28:
                return context$2$0.finish(21);

              case 29:

                nodes = nodesTemp;
                hasParents = nodes.length > 0;

              case 31:
              case 'end':
                return context$2$0.stop();
            }
          }, callee$1$0, _this11, [[4, 17, 21, 29], [22,, 24, 28]]);
        });

      case 8:
        if (!hasParents) {
          context$1$0.next = 12;
          break;
        }

        return context$1$0.delegateYield(_loop4(), 't0', 10);

      case 10:
        context$1$0.next = 8;
        break;

      case 12:
        return context$1$0.abrupt('return', _lodash2['default'].includes(parentsAll, descUI1));

      case 13:
      case 'end':
        return context$1$0.stop();
    }
  }, callee$0$0, this);
}));

/*
* Creates a subtree from a flat list of descriptor record UIs (as `@id`s) based on parent-child relationships within the MeSH ontology tree.
*/
MeshTree.prototype.clusterDescUIs = _bluebird2['default'].coroutine(_regeneratorRuntime.mark(function callee$0$0(idArray) {
  var descUIArray, mostSpecificConcept, _iteratorNormalCompletion13, _didIteratorError13, _iteratorError13, _iterator13, _step13, descUI, children, relationships, _iteratorNormalCompletion14, _didIteratorError14, _iteratorError14, _loop5, _iterator14, _step14, _iteratorNormalCompletion15, _didIteratorError15, _iteratorError15, _iterator15, _step15, treeCluster;

  return _regeneratorRuntime.wrap(function callee$0$0$(context$1$0) {
    var _this13 = this;

    while (1) switch (context$1$0.prev = context$1$0.next) {
      case 0:
        treeCluster = function treeCluster(relationsList, parent, tree) {

          if (typeof tree === 'undefined') {
            tree = [];
          }
          if (typeof parent === 'undefined') {
            parent = { '@id': null, 'parent': null };
          }

          var children = _lodash2['default'].filter(relationsList, function (child) {
            return child.parent === parent['@id'];
          });

          if (!_lodash2['default'].isEmpty(children)) {
            if (_lodash2['default'].isNull(parent['@id'])) {
              tree = children;
            } else {
              parent['children'] = children;
            }

            _lodash2['default'].each(children, function (child) {
              treeCluster(relationsList, child);
            });
          }

          return tree;
        };

        if (_lodash2['default'].isArray(idArray)) {
          context$1$0.next = 3;
          break;
        }

        throw new Error('input not an array.');

      case 3:
        descUIArray = idArray.map(function (id) {
          return id.replace(MESH, '');
        });
        mostSpecificConcept = {};
        _iteratorNormalCompletion13 = true;
        _didIteratorError13 = false;
        _iteratorError13 = undefined;
        context$1$0.prev = 8;
        _iterator13 = _getIterator(descUIArray);

      case 10:
        if (_iteratorNormalCompletion13 = (_step13 = _iterator13.next()).done) {
          context$1$0.next = 19;
          break;
        }

        descUI = _step13.value;
        context$1$0.next = 14;
        return this.getChildren({ id: descUI, format: 'mesh' });

      case 14:
        children = context$1$0.sent;

        mostSpecificConcept[descUI] = !children || !children.length;

      case 16:
        _iteratorNormalCompletion13 = true;
        context$1$0.next = 10;
        break;

      case 19:
        context$1$0.next = 25;
        break;

      case 21:
        context$1$0.prev = 21;
        context$1$0.t0 = context$1$0['catch'](8);
        _didIteratorError13 = true;
        _iteratorError13 = context$1$0.t0;

      case 25:
        context$1$0.prev = 25;
        context$1$0.prev = 26;

        if (!_iteratorNormalCompletion13 && _iterator13['return']) {
          _iterator13['return']();
        }

      case 28:
        context$1$0.prev = 28;

        if (!_didIteratorError13) {
          context$1$0.next = 31;
          break;
        }

        throw _iteratorError13;

      case 31:
        return context$1$0.finish(28);

      case 32:
        return context$1$0.finish(25);

      case 33:
        relationships = [];
        _iteratorNormalCompletion14 = true;
        _didIteratorError14 = false;
        _iteratorError14 = undefined;
        context$1$0.prev = 37;
        _loop5 = _regeneratorRuntime.mark(function callee$1$0() {
          var descUI, parents, nodes, hasParents, parentsFound, _loop6;

          return _regeneratorRuntime.wrap(function callee$1$0$(context$2$0) {
            var _this12 = this;

            while (1) switch (context$2$0.prev = context$2$0.next) {
              case 0:
                descUI = _step14.value;
                parents = [];
                nodes = [descUI];
                hasParents = true;
                parentsFound = false;
                _loop6 = _regeneratorRuntime.mark(function callee$2$0() {
                  var nodesTemp, node, parentsTemp, parentsInArray;
                  return _regeneratorRuntime.wrap(function callee$2$0$(context$3$0) {
                    while (1) switch (context$3$0.prev = context$3$0.next) {
                      case 0:
                        nodesTemp = [];
                        _iteratorNormalCompletion15 = true;
                        _didIteratorError15 = false;
                        _iteratorError15 = undefined;
                        context$3$0.prev = 4;
                        _iterator15 = _getIterator(nodes);

                      case 6:
                        if (_iteratorNormalCompletion15 = (_step15 = _iterator15.next()).done) {
                          context$3$0.next = 16;
                          break;
                        }

                        node = _step15.value;
                        context$3$0.next = 10;
                        return this.getParents({ id: node, format: 'mesh' });

                      case 10:
                        parentsTemp = context$3$0.sent;
                        parentsInArray = _lodash2['default'].intersection(parentsTemp, descUIArray);

                        if (parentsInArray.length > 0) {
                          _lodash2['default'].each(parentsInArray, function (p) {
                            return parents.push(p);
                          });
                          parentsFound = true;
                        } else {
                          _lodash2['default'].each(parentsTemp, function (p) {
                            return nodesTemp.push(p);
                          });
                        }

                      case 13:
                        _iteratorNormalCompletion15 = true;
                        context$3$0.next = 6;
                        break;

                      case 16:
                        context$3$0.next = 22;
                        break;

                      case 18:
                        context$3$0.prev = 18;
                        context$3$0.t0 = context$3$0['catch'](4);
                        _didIteratorError15 = true;
                        _iteratorError15 = context$3$0.t0;

                      case 22:
                        context$3$0.prev = 22;
                        context$3$0.prev = 23;

                        if (!_iteratorNormalCompletion15 && _iterator15['return']) {
                          _iterator15['return']();
                        }

                      case 25:
                        context$3$0.prev = 25;

                        if (!_didIteratorError15) {
                          context$3$0.next = 28;
                          break;
                        }

                        throw _iteratorError15;

                      case 28:
                        return context$3$0.finish(25);

                      case 29:
                        return context$3$0.finish(22);

                      case 30:

                        nodes = nodesTemp;
                        hasParents = nodesTemp.length > 0;

                      case 32:
                      case 'end':
                        return context$3$0.stop();
                    }
                  }, callee$2$0, _this12, [[4, 18, 22, 30], [23,, 25, 29]]);
                });

              case 6:
                if (!(hasParents && !parentsFound)) {
                  context$2$0.next = 10;
                  break;
                }

                return context$2$0.delegateYield(_loop6(), 't0', 8);

              case 8:
                context$2$0.next = 6;
                break;

              case 10:

                if (parents.length === 0) {

                  relationships.push({
                    '@id': MESH + descUI,
                    'parent': null,
                    'mostSpecificConcept': mostSpecificConcept[descUI]
                  });
                } else {

                  _lodash2['default'].each(parents, function (parent) {
                    relationships.push({
                      '@id': MESH + descUI,
                      'parent': MESH + parent,
                      'mostSpecificConcept': mostSpecificConcept[descUI]
                    });
                  });
                }

              case 11:
              case 'end':
                return context$2$0.stop();
            }
          }, callee$1$0, _this13);
        });
        _iterator14 = _getIterator(descUIArray);

      case 40:
        if (_iteratorNormalCompletion14 = (_step14 = _iterator14.next()).done) {
          context$1$0.next = 45;
          break;
        }

        return context$1$0.delegateYield(_loop5(), 't1', 42);

      case 42:
        _iteratorNormalCompletion14 = true;
        context$1$0.next = 40;
        break;

      case 45:
        context$1$0.next = 51;
        break;

      case 47:
        context$1$0.prev = 47;
        context$1$0.t2 = context$1$0['catch'](37);
        _didIteratorError14 = true;
        _iteratorError14 = context$1$0.t2;

      case 51:
        context$1$0.prev = 51;
        context$1$0.prev = 52;

        if (!_iteratorNormalCompletion14 && _iterator14['return']) {
          _iterator14['return']();
        }

      case 54:
        context$1$0.prev = 54;

        if (!_didIteratorError14) {
          context$1$0.next = 57;
          break;
        }

        throw _iteratorError14;

      case 57:
        return context$1$0.finish(54);

      case 58:
        return context$1$0.finish(51);

      case 59:
        return context$1$0.abrupt('return', treeCluster(relationships));

      case 60:
      case 'end':
        return context$1$0.stop();
    }
  }, callee$0$0, this, [[8, 21, 25, 33], [26,, 28, 32], [37, 47, 51, 59], [52,, 54, 58]]);
}));

/*
* Tests whether a descriptor has pharmacological actions (in other words, if the descriptor is a drug).
* If true, returns array of descUI mappings of the pharmacological action, otherwise returns null.
*/
MeshTree.prototype.getPharmacologicalAction = _bluebird2['default'].coroutine(_regeneratorRuntime.mark(function callee$0$0(opts) {
  var ui, format, result;
  return _regeneratorRuntime.wrap(function callee$0$0$(context$1$0) {
    var _this14 = this;

    while (1) switch (context$1$0.prev = context$1$0.next) {
      case 0:
        opts = opts || {};
        ui = this.formatID(opts.id, 'mesh');
        format = opts.format || 'rdf';
        context$1$0.next = 5;
        return this.dbSearch({
          subject: MESH + ui,
          predicate: MESHV + 'pharmacologicalAction',
          object: this.db.v('descUI')
        }, {});

      case 5:
        result = context$1$0.sent;

        if (!_lodash2['default'].isEmpty(result)) {
          context$1$0.next = 10;
          break;
        }

        return context$1$0.abrupt('return', null);

      case 10:
        return context$1$0.abrupt('return', _lodash2['default'].unique(result.map(function (res) {
          return _this14.formatID(res['descUI'], format);
        })));

      case 11:
      case 'end':
        return context$1$0.stop();
    }
  }, callee$0$0, this);
}));

/*
 * Performs mapping of MeSH concepts onto Schema.org classes
 */
MeshTree.prototype.getSchemaOrgType = _bluebird2['default'].coroutine(_regeneratorRuntime.mark(function callee$0$0(opts) {
  var ui, schemaOrgType, pharmActions, treeNums;
  return _regeneratorRuntime.wrap(function callee$0$0$(context$1$0) {
    while (1) switch (context$1$0.prev = context$1$0.next) {
      case 0:
        opts = opts || {};
        ui = this.formatID(opts.id, 'mesh');
        schemaOrgType = 'MedicalEntity';
        context$1$0.next = 5;
        return this.getPharmacologicalAction({ id: ui, format: 'mesh' });

      case 5:
        pharmActions = context$1$0.sent;

        if (pharmActions) {
          schemaOrgType = 'Drug';
        }

        context$1$0.next = 9;
        return this.getTreeNumbers({ id: ui, format: 'mesh' });

      case 9:
        treeNums = context$1$0.sent;

        if (treeNums.some(function (tn) {
          return (/^C\d+/.test(tn)
          );
        })) {
          schemaOrgType = 'MedicalCondition';
          if (treeNums.some(function (tn) {
            return (/^(C01|C02|C03)\./.test(tn)
            );
          })) {
            schemaOrgType = 'InfectiousDisease';
          } else if (treeNums.some(function (tn) {
            return (/^C23\./.test(tn)
            );
          })) {
            schemaOrgType = 'MedicalSignOrSymptom';
          }
        } else if (treeNums.some(function (tn) {
          return (/^E07\.(101|132|190|222|230|278|315|325|430|505|515|605|652|695|814|858|862|877|913|926|935|950)\./.test(tn)
          );
        })) {
          schemaOrgType = 'MedicalDevice';
        } else if (treeNums.some(function (tn) {
          return (/^E01\.370\./.test(tn)
          );
        })) {
          schemaOrgType = 'MedicalTest';
          if (treeNums.some(function (tn) {
            return (/^E01\.370\.(100|370.050|372.250|372.255|372.310|376.300|376.525|376.700|378.150|378.155|378.330|386.105|386.460|386.700|388.100|388.250|390.800|405|530)/.test(tn)
            );
          })) {
            schemaOrgType = 'DiagnosticProcedure';
          } else if (treeNums.some(function (tn) {
            return (/^E01\.370\.(049|350)/.test(tn)
            );
          })) {
            schemaOrgType = 'ImagingTest';
          }
          // skipping BloodTest and PathologyTest mapping for now as these are somewhat ambiguous
        } else if (treeNums.some(function (tn) {
            return (/^E02\./.test(tn)
            );
          })) {
            schemaOrgType = 'MedicalTherapy';
            if (treeNums.some(function (tn) {
              return (/^E02\.815\./.test(tn)
              );
            })) {
              schemaOrgType = 'RadiationTherapy';
            } else if (treeNums.some(function (tn) {
              return (/^E02\.779\./.test(tn)
              );
            })) {
              schemaOrgType = 'PhysicalTherapy';
            } else if (treeNums.some(function (tn) {
              return (/^E02\.(037|065|120|148|154|218|258|278|309|342|365|393|467|514|520|533|565|583|585|594|596|600|621|631|706|718|730|774|794|800|831|870|875|880|891|912|926|950|960)/.test(tn)
              );
            })) {
              schemaOrgType = 'TherapeuticProcedure';
            }
          } else if (treeNums.some(function (tn) {
            return (/^(E03|E04)/.test(tn)
            );
          })) {
            schemaOrgType = 'MedicalProcedure';
          } else if (treeNums.some(function (tn) {
            return (/^F04\.754\./.test(tn)
            );
          })) {
            schemaOrgType = 'PsychologicalTreatment';
          }

        return context$1$0.abrupt('return', schemaOrgType);

      case 12:
      case 'end':
        return context$1$0.stop();
    }
  }, callee$0$0, this);
}));

/*
* Creates properties object from descriptor id
*/
MeshTree.prototype.createPropertiesObject = _bluebird2['default'].coroutine(_regeneratorRuntime.mark(function callee$0$0(propRequestObj) {
  var id, properties, ui, propertiesObj, _iteratorNormalCompletion16, _didIteratorError16, _iteratorError16, _iterator16, _step16, property, preferredTerm, scopeNotes, synonyms, preferredTermIndex, schemaOrgType;

  return _regeneratorRuntime.wrap(function callee$0$0$(context$1$0) {
    while (1) switch (context$1$0.prev = context$1$0.next) {
      case 0:
        id = propRequestObj['@id'];
        properties = propRequestObj.properties;
        ui = id.replace(MESH, '');
        propertiesObj = {
          '@id': id
        };
        _iteratorNormalCompletion16 = true;
        _didIteratorError16 = false;
        _iteratorError16 = undefined;
        context$1$0.prev = 7;
        _iterator16 = _getIterator(properties);

      case 9:
        if (_iteratorNormalCompletion16 = (_step16 = _iterator16.next()).done) {
          context$1$0.next = 48;
          break;
        }

        property = _step16.value;
        preferredTerm = undefined;
        context$1$0.t0 = property;
        context$1$0.next = context$1$0.t0 === 'name' ? 15 : context$1$0.t0 === 'description' ? 20 : context$1$0.t0 === 'synonyms' ? 25 : context$1$0.t0 === 'schemaOrgType' ? 35 : context$1$0.t0 === 'codeValue' ? 40 : context$1$0.t0 === 'codingSystem' ? 42 : 44;
        break;

      case 15:
        context$1$0.next = 17;
        return this.getPrefTerm({ id: ui });

      case 17:
        preferredTerm = context$1$0.sent;

        propertiesObj[property] = preferredTerm;
        return context$1$0.abrupt('break', 45);

      case 20:
        context$1$0.next = 22;
        return this.getScopeNote({ id: ui });

      case 22:
        scopeNotes = context$1$0.sent;

        propertiesObj[property] = scopeNotes;
        return context$1$0.abrupt('break', 45);

      case 25:
        context$1$0.next = 27;
        return this.getPrefTerm({ id: ui });

      case 27:
        preferredTerm = context$1$0.sent;
        context$1$0.next = 30;
        return this.getAllTerms({ id: ui });

      case 30:
        synonyms = context$1$0.sent;
        preferredTermIndex = synonyms.indexOf(preferredTerm);

        if (~preferredTermIndex) synonyms.splice(preferredTermIndex, 1);
        propertiesObj[property] = synonyms;
        return context$1$0.abrupt('break', 45);

      case 35:
        context$1$0.next = 37;
        return this.getSchemaOrgType({ id: ui });

      case 37:
        schemaOrgType = context$1$0.sent;

        propertiesObj[property] = schemaOrgType;
        return context$1$0.abrupt('break', 45);

      case 40:
        propertiesObj[property] = ui;
        return context$1$0.abrupt('break', 45);

      case 42:
        propertiesObj[property] = 'MeSH';
        return context$1$0.abrupt('break', 45);

      case 44:
        propertiesObj[property] = null;

      case 45:
        _iteratorNormalCompletion16 = true;
        context$1$0.next = 9;
        break;

      case 48:
        context$1$0.next = 54;
        break;

      case 50:
        context$1$0.prev = 50;
        context$1$0.t1 = context$1$0['catch'](7);
        _didIteratorError16 = true;
        _iteratorError16 = context$1$0.t1;

      case 54:
        context$1$0.prev = 54;
        context$1$0.prev = 55;

        if (!_iteratorNormalCompletion16 && _iterator16['return']) {
          _iterator16['return']();
        }

      case 57:
        context$1$0.prev = 57;

        if (!_didIteratorError16) {
          context$1$0.next = 60;
          break;
        }

        throw _iteratorError16;

      case 60:
        return context$1$0.finish(57);

      case 61:
        return context$1$0.finish(54);

      case 62:
        return context$1$0.abrupt('return', propertiesObj);

      case 63:
      case 'end':
        return context$1$0.stop();
    }
  }, callee$0$0, this, [[7, 50, 54, 62], [55,, 57, 61]]);
}));

exports['default'] = MeshTree;
module.exports = exports['default'];

// get most represented category if multiply present

// descriptor record UI

// supplemetal record UI

// get rid of descriptor qualifiers as well

// should not be ancestor of common ancestor already in common ancestors array

// remove those in common ancestors array that are broader than current

// input must be array

// booleans whether a concept can have potential children (i.e., not the most specific concept)

// create array of parent-child relationship objects

// recursive function for clustering